use std::collections::HashMap;
use std::sync::mpsc::Sender;
use std::sync::Arc;

use axum::extract::{Path, Query, State};
use axum::http::StatusCode;
use axum::response::IntoResponse;
use axum::routing::get;
use axum::{Json, Router};
use color_eyre::eyre::Context as _;
use color_eyre::Result;
use tokio::net::TcpListener;
use tokio::runtime;

use crate::types::{Event, WebhookEvent};

struct AppState {
    event_tx: Sender<Event>,
}

pub fn worker(event_tx: Sender<Event>) -> Result<()> {
    let rt = runtime::Builder::new_current_thread()
        .thread_name("server")
        .enable_io()
        .build()?;

    rt.block_on(task(event_tx))
}

async fn task(event_tx: Sender<Event>) -> Result<()> {
    let state = AppState { event_tx };
    let shared_state = Arc::new(state);

    let app = Router::new()
        .route(
            "/webhook/:topic",
            get(webhook_handler).post(webhook_handler),
        )
        .with_state(shared_state);

    let listener = TcpListener::bind("0.0.0.0:3000")
        .await
        .wrap_err("Failed to bind to TCP socket")?;
    tracing::info!("Listening on \"0.0.0.0:3000\"");

    axum::serve(listener, app)
        .await
        .wrap_err("Failed to start server")
}

async fn webhook_handler(
    State(state): State<Arc<AppState>>,
    Path(topic): Path<String>,
    Query(query): Query<HashMap<String, String>>,
    body: Option<Json<serde_json::Value>>,
) -> impl IntoResponse {
    let event = Event::Webhook(WebhookEvent {
        topic,
        query,
        body: body
            .map(|Json(body)| body)
            .unwrap_or(serde_json::Value::Null),
    });
    tracing::debug!("Received webhook event: {:?}", &event);

    if state.event_tx.send(event).is_err() {
        tracing::error!("Failed to trigger webhook event");

        (
            StatusCode::INTERNAL_SERVER_ERROR,
            "Failed to trigger webhook event",
        )
    } else {
        (StatusCode::NO_CONTENT, "")
    }
}
